/*
libsndfilegpĂ܂B
libsndfileGNU Lesser General Public Licenseɏ]ĔzzĂ܂B
http://www.mega-nerd.com/libsndfile/
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

/*
* libsndfile is available from http://www.mega-nerd.com/
*/
#include <sndfile.h>


int debugmode = 0;
#define debug_printf(x) {if(debugmode){printf x ;}}

int intget(char *p, int min, int max)
{
	int temp;
	if (sscanf(p, "%d", &temp) != 1) {
		printf("not a number\n");
		exit(-1);
	}
	if (temp < min || max < temp) {
		printf("number out of range\n");
		exit(-1);
	}
	return temp;
}

typedef struct conv_cfg_t {
	long samples;
	int bits;
	int signed_flag;
	int mixmode;
	void (*format)(struct conv_cfg_t *cfg, float *source);
	char *separator;
	int items_per_line;
	void (*convert)(struct conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source);
	char *outfilename;
} conv_cfg_t;


void converter_0(conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source);
void converter_1(conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source);


void (*const converter_ary[])(conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source) = {
	converter_0, converter_1
};

void formatter_0(conv_cfg_t *cfg, float *source);
void formatter_1(conv_cfg_t *cfg, float *source);
void formatter_2(conv_cfg_t *cfg, float *source);
void formatter_3(conv_cfg_t *cfg, float *source);


void (*const formatter_ary[])(conv_cfg_t *cfg, float *source) = {
	formatter_0, formatter_1, formatter_2, formatter_3
};

void cfg_t_init(conv_cfg_t *cfg)
{
	cfg->samples = 0;
	cfg->bits = 0;
	cfg->signed_flag = 0;
	cfg->mixmode = 2;
	cfg->format = formatter_ary[0];
	cfg->separator = ", ";
	cfg->items_per_line = 8;
	cfg->convert = converter_ary[1];
	cfg->outfilename = NULL;
}

void usage(void)
{
	printf("Usage: sf2wt [OPTION]... [SOURCEFILE] [OUTPUTFILE]\n");
	printf("if OUTPUTFILE is not specified then stdout is used.\n");
	printf("\n");
	printf("convert options:\n");
	printf("   -n?    number of output sample\n");
	printf("   -b?    quantization bit rate(1-8)\n");
	printf("   -s     output data is signed\n");
	printf("   -u     output data is unsigned(default)\n");
	printf("   -c?    0:L 1:R 2:mix if source is stereo \n");
	printf("   -d?    algorithm select\n");
	printf("          0:simple 1:average(default)\n");
	//printf("   -i?    algorithm select\n");
	printf("format options:\n");
	printf("   -f?    format select\n");
	printf("          0:decimal, 1:$xx, 2:0xxx, 3:binary\n");
	printf("   -l?    items per line \n");
	printf("   -j?    item seperator\n");
	printf("\n");
	printf("example for FDS:\n");
	printf("   sf2wt -n64 -b6 -u -l8 source.wav\n");
	printf("\n");
	printf("example for HES:\n");
	printf("   sf2wt -n32 -b5 -u -l8 source.wav\n");
	printf("\n");
}


int main(int argc, char *argv[])
{
	char *infname = NULL;
	SNDFILE *sndfile = NULL;
	SF_INFO sfinfo;
	float *source = NULL, *dest = NULL;
	conv_cfg_t cfg;
	int i;
	cfg_t_init(&cfg);
	
	
	if (argc < 2) {
		usage();
		return 0;
	}
	for (i = 1; argv[i] != NULL; i++) {
		char *p = argv[i];
		if (*p++ == '-') {
			switch (*p++) {
			case 'h':
				usage();
				return 0;
			case 'n':
				cfg.samples = intget(p, 2, INT_MAX);
				break;
			case 'b':
				cfg.bits = intget(p, 1, 8);
				break;
			case 's':
				cfg.signed_flag = 1;
				break;
			case 'u':
				cfg.signed_flag = 0;
				break;
			case 'c':
				cfg.mixmode = intget(p, 0, 2);
				break;
			case 'l':
				cfg.items_per_line = intget(p, 1, INT_MAX);
				break;
			case 'j':
				switch (intget(p, 0, 2)) {
				case 0:
					cfg.separator = " ";
					break;
				case 1:
					cfg.separator = ", ";
					break;
				case 2:
					cfg.separator = ",";
					break;
				}
				break;
			case 'f':
				cfg.format = formatter_ary[intget(p, 0, -1 + sizeof(formatter_ary) / sizeof(formatter_ary[0]))];
				break;
			case 'd':
				cfg.convert = converter_ary[intget(p, 0, -1 + sizeof(converter_ary) / sizeof(converter_ary[0]))];
				break;
			case '\0':
				printf("invalid option -\n");
				usage();
				return -1;
			default:
				printf("invalid option -%c\n", *--p);
				usage();
				return -1;
			}
		} else {
			if (infname != NULL) {
				if (cfg.outfilename != NULL) {
					usage();
					return -1;
				} else {
					cfg.outfilename = argv[i];
				}
			} else {
				infname = argv[i];
			}
		}
	}
	
	if (infname == NULL) {
		printf("file not specified\n");
		return -1;
	}
	if (cfg.samples == 0 || cfg.bits == 0) {
		printf("convert option not specified\n");
		printf("-b and -n are needed at least\n");
		return -1;
	}

	
	sfinfo.format = 0;
	
	sndfile = sf_open(infname, SFM_READ, &sfinfo);
	if (!sndfile) {
		printf("can't open %s\n", infname);
		printf("%s\n", sf_strerror(sndfile));
		return -1;
	}

	if (sfinfo.channels != 1 && sfinfo.channels != 2) {
		printf("can't convert from source which has %d channels\n", sfinfo.channels);
		goto on_error;
	}

	if (sfinfo.frames == 0) {
		printf("no samples\n");
		goto on_error;
	}
	
	printf("source samples %ld\n", (long)sfinfo.frames);
	printf("out samples %ld\n", cfg.samples);
	{
		double ratio;
		ratio = (double)cfg.samples/(double)sfinfo.frames;
		printf("rate %.10lf\n", ratio);
		if (ratio > 1.0) {
			//printf(": won't do any interpolation ԉ܂in upsampling\n");
			cfg.convert = converter_ary[0];
			//goto on_error;
		}
	}
	if (cfg.signed_flag) {
		printf("out data range %d to %d\n", -1 * (1<<(cfg.bits-1)), (1<<(cfg.bits-1)) - 1);
	} else {
		printf("out data range %d to %d\n", 0, (1<<(cfg.bits)) - 1);
	}
	printf("\n");
	
	source = (float *)malloc(sfinfo.frames * sfinfo.channels * sizeof(*source));
	if (source == NULL) {
		printf("out of memory\n");
		goto on_error;
	}
	if (sfinfo.frames != sf_readf_float(sndfile, source, sfinfo.frames)) {
		printf("read error\n");
		goto on_error;
	}
	
	//mix or extract channels
	if (sfinfo.channels == 2) {
		for (i = 0; i < sfinfo.frames; i++) {
			switch (cfg.mixmode) {
			case 0:
				source[i] = source[i * sfinfo.channels];
				break;
			case 1:
				source[i] = source[i * sfinfo.channels + 1];
				break;
			case 2:
				source[i] = source[i * sfinfo.channels] + source[i * sfinfo.channels + 1];
				source[i] /= 2.0;
				break;
			}
		}
	}
	
	for (i = 0; i < sfinfo.frames;i++) {
		debug_printf(("s %f\n", source[i]));
	}
	
	dest = (float *)malloc(cfg.samples * sizeof(*dest));
	if (dest == NULL) {
		printf("outofmemory\n");
		goto on_error;
	}
	
	{	
		float max = -1.0, min = 1.0, nommax = 0.0, zero = 0.0;
		float digitize_rate;
		float offset = 0.0;
		
		cfg.convert(&cfg, &sfinfo, dest, source);
		
		/*
		m[}CY
		 5    ...           max = 5  +
		 4   .   .                   |̒ 5 - 23
		 3  .     .        .         |
		 2 .       .      . zero = 2 +    zero = (5 + -1) /2
		 1.         .    .           |̒ -1 - 23
		 0-----------.--.---0        |
		-1            ..    min = -1 +
		
		*/
		for (i = 0; i < cfg.samples;i++) {
			debug_printf(("%f\n", dest[i]));
			max = dest[i] > max ? dest[i] : max;
			min = dest[i] < min ? dest[i] : min;
		}
		zero = (max + min)/2.0; //vXɂĂvX
		nommax = max - zero;// == zero - min
		
		debug_printf(("dest max %f min %f offset %f normalmax %f\n", max, min, zero, nommax));
		
		for (i = 0; i < cfg.samples; i++) {
			dest[i] -= zero;
			dest[i] /= nommax;
			debug_printf(("%10.10f\n", dest[i]));
		}
		debug_printf(("\n"));
		
		//[-1.0 -- 1.0] -> [min -- max)
		//[-1.0 -- 1.0] -> [0 -- 16)̂΂(4bit)
		//8΂8[0.0 -- 16.0]8=1<<3
		digitize_rate = (float)(1<<(cfg.bits-1));
		//ƂƂ(1.0҂ɂȂ16ɂȂĂ4bitł͂Ȃ̂)
		digitize_rate *= 0.9999;
		if (cfg.signed_flag) {
			//t̏ꍇ̓Q^
			offset = 0.0;
		} else {
			offset = 1.0;
		}
		
		for (i = 0; i < cfg.samples; i++) {
			dest[i] = (dest[i] + offset) * digitize_rate;
			debug_printf(("%2d    ", (int)dest[i]));
			debug_printf(("%.2f", dest[i]));
			debug_printf(("\n"));
		}
		cfg.format(&cfg, dest);
	}
	
	
	
	
	free(dest);
	free(source);
	sf_close(sndfile);
	return 0;
on_error:
	if (dest) {
		free(dest);
		source = NULL;
	}
	if (source) {
		free(source);
		source = NULL;
	}
	if (sndfile) {
		sf_close(sndfile);
		sndfile = NULL;
	}
	return -1;
}


/*
Ƃ肾
300->10̂΂
0 0
1 30
2 60
3 90
4 120
5 150
6 180
7 210
8 240
9 270
*/
void converter_0(conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source)
{
	int i, j;
	double d = 0.0;
	double ratio = (double)sfinfo->frames / (double)cfg->samples;
	for (i= 0; i < cfg->samples; i++) {
		d = (double) i * ratio;
		j = (int)(d + 0.5);
		if (j >= sfinfo->frames) {
			j = sfinfo->frames - 1;
		}
		//printf("s %d -> d %d\n", j, i);
		dest[i] = source[j];
		
	}
}


/*
ړςۂ
300->10̂΂
0 0`29average
1 30
2 60
3 90
4 120
5 150
6 180
7 210
8 240
9 270`299
*/
void converter_1(conv_cfg_t *cfg, SF_INFO *sfinfo, float *dest, float *source)
{
	int i, j, kono_mae_made = 0, kokokara = 0;
	double ratio = (double)sfinfo->frames / (double)cfg->samples;
	for (i= 0; i < cfg->samples; i++) {
		kokokara = (int)((double)(i) * ratio + 0.5);
		if (kokokara > sfinfo->frames - 1) {
			kokokara = sfinfo->frames - 1;
		}
		kono_mae_made = (int)((double)(i + 1) * ratio + 0.5);
		if (kono_mae_made > sfinfo->frames) {
			kono_mae_made = sfinfo->frames;
		}
		
		if (kokokara != kono_mae_made) {
			dest[i] = 0.0;
			for (j = kokokara; j < kono_mae_made; j++) {
				dest[i] += source[j];
			}
			dest[i] /= (kono_mae_made - kokokara);
		} else {
			//AbvTvÔƂ
			dest[i] = source[kokokara];
		}
		
		debug_printf(("s %d -- %d (%d)-> d %d    %f\n", kokokara, kono_mae_made, kono_mae_made - kokokara, i, dest[i]));
	}
}







void text_formatter(conv_cfg_t *cfg, float *source, char *strintform, int compl_flag)
{
	int i = 0, byte;
	FILE *fp = NULL;
	int file_opened = 0;
	if (cfg->outfilename) {
		fp = fopen(cfg->outfilename, "wt");
		if (!fp) {
			printf("can't open %s\n", cfg->outfilename);
			fp = stdout;
		} else {
			file_opened = 1;
		}
	} else {
		fp = stdout;
	}
	
	while (1) {
		byte = (int)floorf(source[i]);
		if (cfg->signed_flag && byte < 0 && compl_flag) {
			byte &= 0xff;
		}
		fprintf(fp, strintform, byte);
		
		/* loop check */
		if (++i >= cfg->samples) {
			break;
		}
		fprintf(fp, "%s", cfg->separator);
		if (i % cfg->items_per_line == 0) {
			fprintf(fp, "\n");
		}
	}
	fprintf(fp, "\n");
	
	if (file_opened && fp != NULL && fp != stdout) {
		fclose(fp);
	}
}



/* decimal */
void formatter_0(conv_cfg_t *cfg, float *source)
{
	text_formatter(cfg, source, "%4d", 0);
}

/* hex ($xx) */
void formatter_1(conv_cfg_t *cfg, float *source)
{
	text_formatter(cfg, source, "$%02x", 1);
}

/* hex (0x??) */
void formatter_2(conv_cfg_t *cfg, float *source)
{
	text_formatter(cfg, source, "0x%02x", 1);
}

/* binary output */
void formatter_3(conv_cfg_t *cfg, float *source)
{
	int i, byte;
	FILE *fp = NULL;
	if (cfg->outfilename) {
		fp = fopen(cfg->outfilename, "wb");
		if (!fp) {
			printf("can't open %s\n", cfg->outfilename);
			return;
		}
	} else {
		printf("output file not specified in binary output mode\n");
		return;
	}
	
	for (i = 0; i < cfg->samples; i++) {
		byte = (int)floorf(source[i]);
		if (cfg->signed_flag && byte < 0) {
			byte &= 0xff;
		}
		fprintf(fp, "%c", byte);
	}
	
	fclose(fp);
}





